from django.conf import settings
from django.db import models
from django.utils import timezone
from ipware import get_client_ip

from vnpay.utils import VnPay

STATUS_CHOICES = (
    ('NEW', 'NEW'),
    ('CONFIRMED', 'CONFIRMED'),
)

RESULT_CHOICES = (
    ('VNPAY_00', 'VNPAY SUCCESS'),
    ('VNPAY_05', 'VNPAY BAD_PASSWORD_EXCEEDING'),
    ('VNPAY_06', 'VNPAY WRONG_PASSWORD'),
    ('VNPAY_07', 'VNPAY SUSPICIOUS_TRANSACTION'),
    ('VNPAY_12', 'VNPAY CARD_IS_LOCKED'),
    ('VNPAY_09', 'VNPAY NO_INTERNET_BANKING'),
    ('VNPAY_10', 'VNPAY VERIFICATION_EXCEEDING'),
    ('VNPAY_11', 'VNPAY TIMEOUT_PAYMENT'),
    ('VNPAY_24', 'VNPAY CUSTOMER_CANCEL'),
    ('VNPAY_51', 'VNPAY NOT_ENOUGH_BALANCE'),
    ('VNPAY_65', 'VNPAY TRANSACTION_PER_DAY_EXCEEDING'),
    ('VNPAY_75', 'VNPAY BANK_MAINTENANCE'),
    ('VNPAY_99', 'VNPAY OTHER_ERRORS'),
)


class Billing(models.Model):
    created_at = models.DateTimeField(default=timezone.now)
    updated_at = models.DateTimeField(auto_now=True)

    amount = models.DecimalField(max_digits=10, decimal_places=2)
    currency = models.CharField(default='VND', max_length=3)
    pay_at = models.DateTimeField(null=True, blank=True)
    pay_by = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, null=True)
    reference_number = models.CharField(
        max_length=100, help_text='Reference number from payment gateway', blank=True,
        null=True,
    )

    status = models.CharField(choices=STATUS_CHOICES, default='NEW', max_length=20)
    result_payment = models.CharField(choices=RESULT_CHOICES, max_length=50, null=True)
    is_paid = models.BooleanField(default=False)
    transaction_id = models.CharField(max_length=255, null=True, blank=True)

    def set_paid(self):
        self.result_payment = 'succeeded'
        self.result_payment = 'VNPAY_00'
        self.pay_at = timezone.now()
        self.is_paid = True

    def finalize(self, succeeded):
        if not succeeded:
            return
        self.set_paid()
        self.save()

    def get_payment_url(self, request):
        ipaddr, is_routable = get_client_ip(request)

        vnp = VnPay()
        request_data = {
            'vnp_Version': '2.1.0',
            'vnp_Command': 'pay',
            'vnp_TmnCode': settings.VNPAY_TMN_CODE,
            'vnp_Amount': int(self.amount * 100),
            'vnp_CurrCode': 'VND',
            'vnp_TxnRef': self.reference_number,
            'vnp_OrderInfo': f'Thanh toán hóa đơn {self.reference_number}',
            'vnp_OrderType': 'billpayment',
            'vnp_Locale': 'vn',
            'vnp_CreateDate': timezone.now().strftime('%Y%m%d%H%M%S'),
            'vnp_IpAddr': ipaddr,
            'vnp_ReturnUrl': settings.VNPAY_RETURN_URL,
        }
        return vnp.get_payment_url(request_data)



import hashlib
import hmac
from urllib.parse import quote_plus

from django.conf import settings
from rest_framework import serializers


# https://sandbox.vnpayment.vn/apis/docs/thanh-toan-pay/pay.html#code-ipn-url
class VnPaySerializer(serializers.Serializer):
    vnp_TmnCode = serializers.CharField()
    vnp_Amount = serializers.CharField()
    vnp_BankCode = serializers.CharField()
    vnp_BankTranNo = serializers.CharField(required=False)
    vnp_CardType = serializers.CharField(required=False)
    vnp_PayDate = serializers.CharField(required=False)
    vnp_OrderInfo = serializers.CharField()
    vnp_TransactionNo = serializers.CharField()
    vnp_ResponseCode = serializers.CharField()
    vnp_TransactionStatus = serializers.CharField()
    vnp_TxnRef = serializers.CharField()
    vnp_SecureHashType = serializers.CharField(required=False)
    vnp_SecureHash = serializers.CharField()


def hmacsha512(key, data):
    byteKey = key.encode('utf-8')
    byteData = data.encode('utf-8')
    return hmac.new(byteKey, byteData, hashlib.sha512).hexdigest()


class VnPay:
    requestData = {}

    def get_query_string_and_hash(self, data):
        input_data = sorted(data.items())
        query_string = ''
        seq = 0
        for key, val in input_data:
            if seq == 1:
                query_string += f'&{key}={quote_plus(str(val))}'
            else:
                seq = 1
                query_string = f'{key}={quote_plus(str(val))}'

        hash_value = hmacsha512(settings.VNPAY_HASH_SECRET_KEY, query_string)
        return query_string, hash_value

    def get_payment_url(self, request_data):
        query_string, hash_value = self.get_query_string_and_hash(request_data)
        return f'{settings.VNPAY_PAYMENT_URL}?{query_string}&vnp_SecureHash={hash_value}'

    def validate_response_data(self, response_data):
        serializers = VnPaySerializer(data=response_data)
        serializers.is_valid(raise_exception=True)
        return serializers.validated_data

    def validate_hash(self, response_data):
        vnp_secure_hash = response_data.pop('vnp_SecureHash')

        if 'vnp_SecureHashType' in response_data.keys():
            response_data.pop('vnp_SecureHashType')

        _, hash_value = self.get_query_string_and_hash(response_data)
        return vnp_secure_hash == hash_value



from django.conf import settings
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APITestCase

from accounts.models import User
from vnpay.models import Billing
from vnpay.utils import hmacsha512


class BaseUserTest(APITestCase):
    def setUp(self):
        self.user = User.objects.create_user(
            username='test',
            password='test',
        )
        self.client.force_authenticate(self.user)


class BaseBilling(BaseUserTest):
    def setUp(self):
        super().setUp()
        self.billing = Billing.objects.create(
            pay_by=self.user,
            status='NEW',
            amount=100000,
            reference_number='00001111',
        )

    def _get_query_string(self, vnp_Amount=10000000, vnp_ResponseCode='00', vnp_TxnRef='00001111'):  # default success
        return f'vnp_Amount={vnp_Amount}' \
               '&vnp_BankCode=NCB' \
               '&vnp_BankTranNo=VNP12345678' \
               '&vnp_CardType=ATM' \
               f'&vnp_OrderInfo=Thanh+toan+vnpay+hoa+don+{vnp_TxnRef}' \
               '&vnp_PayDate=20230328175935' \
               f'&vnp_ResponseCode={vnp_ResponseCode}' \
               '&vnp_TmnCode=EOH00000' \
               '&vnp_TransactionNo=12345678' \
               '&vnp_TransactionStatus=00' \
               f'&vnp_TxnRef={vnp_TxnRef}'

    def _call_api(self, query_string, has_secure_type=False):
        hash_result = hmacsha512(settings.VNPAY_HASH_SECRET_KEY, query_string)
        url = f'{self.url}?{query_string}&vnp_SecureHash={hash_result}'

        if has_secure_type:
            url += '&vnp_SecureHashType=SHA256'

        return self.client.get(url)


class VnPayPaymentUrlApiTest(BaseUserTest):
    def test_payment_url(self):
        url = reverse('payment-url')
        data = {
            'amount': 1000000,
        }
        response = self.client.post(url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('payment_url', response.data)
        self.assertTrue(response.data['payment_url'])

        self.assertTrue(Billing.objects.filter(pay_by=self.user).exists())
        billing = Billing.objects.get(pay_by=self.user)

        self.assertEqual(billing.amount, data['amount'])
        self.assertEqual(billing.currency, 'VND')
        self.assertEqual(billing.status, 'NEW')

        self.assertTrue(billing.reference_number)
        self.assertTrue(billing.created_at)
        self.assertTrue(billing.updated_at)
        self.assertTrue(billing.id)


class VnPayPaymentIpnApiTests(BaseBilling):
    url = reverse('payment-ipn')

    def _assert_response_success(self, response, assert_response):
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, assert_response)

    def test_vnpay_payment_ipn_view_get(self):
        response = self.client.get(self.url)
        self._assert_response_success(response, {'RspCode': '99', 'Message': 'Invalid request'})

    def test_vnpay_payment_ipn_wrong_secret_key(self):
        query_string = self._get_query_string()
        response = self.client.get(f'{self.url}?{query_string}&vnp_SecureHash=wrong_hash_key&vnp_SecureHashType=SHA256')
        self._assert_response_success(response, {'RspCode': '97', 'Message': 'Invalid Signature'})

    def test_vnpay_payment_ipn_right_secret_key_billing_updated(self):
        self.billing.status = 'CONFIRMED'
        self.billing.save()

        query_string = self._get_query_string()
        response = self._call_api(query_string)
        self._assert_response_success(
            response, {'RspCode': '02', 'Message': 'Order Already Update'},
        )

    def test_vnpay_payment_ipn_success(self):
        query_string = self._get_query_string()
        response = self._call_api(query_string)
        self._assert_response_success(response, {'RspCode': '00', 'Message': 'Confirm Success'})

        self.billing.refresh_from_db()
        self.assertEqual(self.billing.result_payment, 'VNPAY_00')

    def test_vnpay_payment_ipn_success_but_payment_not_success(self):
        query_string = self._get_query_string(vnp_ResponseCode='99')
        response = self._call_api(query_string)
        self._assert_response_success(response, {'RspCode': '00', 'Message': 'Confirm Success'})

        self.billing.refresh_from_db()
        self.assertEqual(self.billing.result_payment, 'VNPAY_99')

    def test_vnpay_payment_ipn_order_not_found(self):
        query_string = self._get_query_string(vnp_TxnRef='999999999')
        response = self._call_api(query_string)
        self._assert_response_success(response, {'RspCode': '01', 'Message': 'Order not found'})

    def test_vnpay_payment_ipn_invalid_amount(self):
        query_string = self._get_query_string(vnp_Amount=1000)
        response = self._call_api(query_string)
        self._assert_response_success(response, {'RspCode': '04', 'Message': 'Invalid Amount'})


class VnPayPaymentReturnApiTests(BaseBilling):
    url = reverse('payment-return')

    def test_vnpay_payment_return_view_get(self):
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)

    def test_vnpay_payment_return_right_secret_key_payment_success(self):
        query_string = self._get_query_string()
        response = self._call_api(query_string)
        self.assertEqual(response.status_code, 200)
        self.assertIn('- Thành công'.encode(), response.content)

    def test_vnpay_payment_has_secure_type(self):
        query_string = self._get_query_string()
        response = self._call_api(query_string, True)
        self.assertEqual(response.status_code, 200)
        self.assertIn('- Thành công'.encode(), response.content)

    def test_vnpay_payment_return_right_secrect_key_payment_failed(self):
        query_string = self._get_query_string(vnp_ResponseCode='99')
        response = self._call_api(query_string)
        self.assertEqual(response.status_code, 200)
        self.assertIn('Không thành công'.encode(), response.content)

    def test_vnpay_payment_return_wrong_checksum(self):
        query_string = self._get_query_string(vnp_ResponseCode='00')
        url = f'{self.url}?{query_string}&vnp_SecureHash=123'

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertIn('Không thành công'.encode(), response.content)
        self.assertIn('Sai checksum'.encode(), response.content)
